from collections import defaultdict
from rest_framework import serializers
from django.db import models, connection
from django.db import ProgrammingError
from django.db.models.fields import related_descriptors
from .ts import export_interface
from .exceptions import UnknownProperty
from .related_properties import is_related_property, get_accessor, get_reverse_accessor

class StateModel:
    """The StateModel is constructed based on a nested serializers.ModelSerializer
    instance. It introspects the serializer and recursively build a model with all
    layers of the serializer.
    """

    def __init__(self, state_serializer, many=False, origin=None, instance_property=None, query_property=None, reverse_acessor=None):
        self.nested_serializer = state_serializer
        self.many = many
        self.origin = origin
        self.reverse_acessor = reverse_acessor

        if origin is None:
            # This is the top-most call, this layer is the anchor
            self.anchor = self
            self.instance_path = []
            self.query_path = []
            self.index = defaultdict(list)
        else:
            self.anchor = origin.anchor
            self.instance_path = origin.instance_path[:] + [instance_property]
            self.query_path = origin.query_path[:] + [query_property]
            self.index = origin.index

        try:
            meta = state_serializer.Meta
        except AttributeError:
            meta = state_serializer.child.Meta

        self.model = meta.model

        if origin:
            self.anchor_key = '__'.join(self.query_path)
        else:
            self.anchor_key = self.model._meta.pk.name

        self.instance_type = '.'.join([
            self.nested_serializer.__module__,
            self.nested_serializer.__class__.__name__,
        ])
        #modu = self.nested_serializer.__module__.replace('.serializers', '')
        #print(f"""perl -pi -e "s/instance_type === '{modu}.{self.model.__name__}/instance_type === '{self.instance_type}/" $1""")

        self.index[self.instance_type].append(self)

        self.user_key = getattr(meta, 'user_key', None)
        if self.user_key and self.user_key not in meta.fields:
            raise ProgrammingError(f'{state_serializer.__class__.__name__}.Meta declares user_key="{self.user_key}", but "{self.user_key}" is not in fields')
        self.optimistic = getattr(meta, 'optimistic', False)
        self.optimistic_timeout = getattr(meta, 'optimistic_timeout', 3)

        self.flat_serializer, fields = self._disassemble_nested()

        self.children = {}
        for field_name, serializer in fields.items():
            node = self._build_child(field_name, serializer)
            if node:
                self.children[field_name] = node

        export_interface(self.nested_serializer.__class__)

    def __str__(self):
        return f'StateModel for {self.instance_type}'

    def __repr__(self):
        return str(self)

    def __getitem__(self, key):
        return self.children[key]

    def models(self):
        for models in self.index.values():
            for model in models:
                yield model

    def frontend_model(self):
        frontend = {}
        for key, nodes in self.index.items():
            node = nodes[0]
            instance_model = {}
            frontend[key] = instance_model
            serializer = node.nested_serializer
            for field_name, field in serializer._declared_fields.items():
                if is_model_serializer(field):
                    instance_model[field_name] = node[field_name].instance_type

        return frontend

    def get_anchors(self, serialized):
        """Get all anchors that should receive an instance"""
        peer_type = serialized['_instance_type']
        peer_models = self.index[peer_type]
        for peer_model in peer_models:
            anchor_key = peer_model.anchor_key
            kwargs = {anchor_key: serialized['id']}
            for instance in self.anchor.model.objects.filter(**kwargs):
                yield instance

    def serialize_instance(self, instance, tstamp):
        data = self.flat_serializer(instance).data
        data['_deleted'] = False
        return self._mark(data, tstamp)

    def serialize_delete(self, instance, tstamp):
        pk = instance.pk or instance.id
        data = self._mark({'id': pk}, tstamp)
        data['_deleted'] = True
        data['_operation'] = 'delete'
        return data

    def serialize_state(self, instance, tstamp):
        if self.many:
            data = self.flat_serializer(instance.all(), many=True).data
            instances = instance.all()
        else:
            data = [ self.flat_serializer(instance).data ]
            instances = [instance]

        for serialized in data:
            self._mark(serialized, tstamp)

        yield data

        for field_name, peer_model in self.children.items():
            for instance in instances:
                try:
                    peer_instance = getattr(instance, field_name)
                except AttributeError:
                    continue
                if peer_instance is None:
                    continue
                for serialized in peer_model.serialize_state(peer_instance, tstamp):
                    yield serialized

    def _mark(self, serialized, tstamp):
        serialized['_instance_type'] = self.instance_type
        serialized['_tstamp'] = tstamp
        serialized['_operation'] = 'initial_state'
        if self.user_key:
            serialized['_user_key'] = serialized.get(self.user_key, None)
        else:
            serialized['_user_key'] = None
        return serialized

    def _disassemble_nested(self):
        serializer_fields = {}
        declared_fields = {}

        for field_name in self.nested_serializer.fields.keys():
            field = self.nested_serializer._declared_fields.get(field_name)

            if field is None:
                continue

            if is_related_property(self.model, field_name):
                try:
                    new_field = serializers.PrimaryKeyRelatedField(
                        many=field.many, read_only=True,
                    )
                except AttributeError:
                    new_field = serializers.PrimaryKeyRelatedField(
                        read_only=True,
                    )
                serializer_fields[field_name] = field
                declared_fields[field_name] = new_field
            elif is_model_serializer(field):
                serializer_fields[field_name] = field
            else:
                declared_fields[field_name] = field

        # Return the new flat serializer class and the dictionary of substituted fields
        FlatSerializer = type(
            self.nested_serializer.__class__.__name__,
            (serializers.ModelSerializer,),
            {
                "Meta": self.nested_serializer.Meta,
                **declared_fields,
            }
        )

        return FlatSerializer, serializer_fields

    def _build_child(self, field_name, serializer):
        try:
            descriptor = getattr(self.model, field_name)
        except AttributeError:
            return

        if isinstance(descriptor, related_descriptors.ManyToManyDescriptor):
            related_descriptor = getattr(descriptor.field.model, descriptor.field.name)
            return StateModel(
                serializer.child,
                True,
                self,
                field_name,
                field_name,
                descriptor.field.name,
            )

        if isinstance(descriptor, related_descriptors.ReverseManyToOneDescriptor):
            related_descriptor = getattr(descriptor.field.model, descriptor.field.name)
            return StateModel(
                serializer.child,
                True,
                self,
                field_name,
                related_descriptor.field.related_query_name(),
                descriptor.field.name,
            )

        if isinstance(descriptor, related_descriptors.ForwardManyToOneDescriptor):
            return StateModel(
                serializer,
                False,
                self,
                field_name,
                field_name,
                descriptor.field.related_query_name(),
            )

        if isinstance(descriptor, related_descriptors.ReverseOneToOneDescriptor):
            return StateModel(
                serializer,
                False,
                self,
                field_name,
                field_name,
                descriptor.related.remote_field.name,
            )

        if isinstance(descriptor, property):
            query_property = get_accessor(self.model, field_name)
            reverse_acessor = get_reverse_accessor(self.model, field_name)
            if not query_property:
                my_module = self.__class__.__module__
                raise UnknownProperty(
                    f'Unknown property {field_name} in '
                    f'{self.model.__module__}.{self.model.__name__}.\n'
                    f'Use {my_module}.decorators.related_property '
                    'to provide an acessor'
                )

            many = getattr(serializer, 'many', False)
            return StateModel(
                serializer.child if many else serializer,
                many,
                self,
                field_name,
                query_property,
                reverse_acessor,
            )


def is_model_serializer(field):
    try:
        return isinstance(field.child, serializers.ModelSerializer)
    except AttributeError:
        return isinstance(field, serializers.ModelSerializer)



import json
import channels.layers
from channels.db import database_sync_to_async
from django.db import ProgrammingError
from rest_framework import serializers

from .consumers import StateConsumer, get_consumer_methods
from .state_model import StateModel
from .state_loader import StateLoader
from .websocket_router import WebsocketRouter
from .signal_handler import SignalHandler
from .redis import RedisStateSession
from .mongo import MongoStateSession
from .serialize import json_dumps

class ContextChannelMeta(type):
    """Metaclass for the ContextChannel.

    Builds the state model based on the provided fields and Meta class,
    and assigns a WebsocketRouter and a SignalHandler to it.
    """

    def __new__(cls, name, bases, attrs):
        """Create and return a new class instance."""

        # Create the new class as usual.
        new_class = super().__new__(cls, name, bases, attrs)

        if new_class.__module__ == cls.__module__:
            return new_class

        # Set the name to be used for mongo collection names and redis keys.
        new_class.name = '_'.join([
            new_class.__module__.replace('.channels', '').replace('.', '_'),
            new_class.__name__.lower()
        ])

        # Ensure the Meta class and anchor serializer are provided.
        meta = attrs.get("Meta")
        if not meta:
            raise ProgrammingError(
                f'{new_class.__name__} must define a Meta class'
            )

        meta.abstract = getattr(meta, 'abstract', False)
        if meta.abstract:
            return new_class

        # The "state" property defines the top-level serializer that will
        # be used to build the context state.
        # This top-level serializer is called "anchor" all over the code,
        # while anchor_id refers to the instance used to build the state.
        if not hasattr(meta, "state"):
            raise ProgrammingError(
                f'{new_class.__name__} must define a Meta class setting the'
                '"state" property to a serializer inheriting from'
                'serializers.ModelSerializer'
            )

        anchor = meta.state
        many = False
        if isinstance(anchor, serializers.ListSerializer):
            many = True
            anchor = anchor.child

        meta.auto_update = getattr(meta, 'auto_update', False) and many

        if not isinstance(anchor, serializers.ModelSerializer):
            raise ProgrammingError(
                f'{new_class.__name__}.Meta.state must be an instance of '
                'serializers.ModelSerializer'
            )

        # Attach the state model, websocket router, and signal handler.
        new_class.meta = meta
        new_class.many = many
        new_class._state_model = StateModel(anchor)
        new_class._wsrouter = WebsocketRouter(new_class.name)
        new_class._signal_handler = SignalHandler(new_class)
        new_class._anchor_model = anchor.Meta.model
        new_class._anchor_events_channel = f'{new_class.__name__}-anchor-events'
        new_class._consumer_methods = get_consumer_methods(new_class)
        return new_class


class ContextChannel(metaclass=ContextChannelMeta):
    """This is the core API of RxDjango.
    ContextChannel provides a state that is synchronized with
    all connected clients through a websocket.
    Frontend classes are automatically generated for each
    ContextChannel subclass.
    Subclasses of ContextChannel should reside in a file named
    channels.py inside a django app.
    """

    class Meta:
        abstract = True

    RuntimeState = None

    @classmethod
    def as_asgi(cls):
        Consumer = type(
            f'{cls.__name__}Consumer',
            (StateConsumer,),
            dict(
                context_channel_class=cls,
                wsrouter=cls._wsrouter,
            ),
        )

        return Consumer.as_asgi()

    def __init__(self, user, **kwargs):
        self.kwargs = kwargs
        self.user = user
        self.user_id = user.id
        self._consumer = None  # will be set by consumer
        self.runtime_state = self.RuntimeState() if self.RuntimeState else None
        self.anchor_ids = []
        self.anchor_index = set()

    async def send(self, *args, **kwargs):
        """A proxy method to self._consumer.send"""
        await self._consumer.send(*args, **kwargs)

    async def group_add(self, group):
        """Add this consumer to a group in channels"""
        channel_layer = channels.layers.get_channel_layer()
        await channel_layer.group_add(
            group,
            self._consumer.channel_name,
        )

    async def initialize_anchors(self):
        if self.many:
            qs = await self.list_instances(**self.kwargs)
            self.anchor_ids = await self._fetch_instance_ids(qs)
        else:
            self.anchor_ids = [
                self.get_instance_id(**self.kwargs)
            ]
        self.anchor_index = set(self.anchor_ids)

    @database_sync_to_async
    def _fetch_instance_ids(self, qs):
        return [
            instance['id'] for instance in qs.values('id')
        ]

    def get_instance_id(self, **kwargs):
        """Subclass may implement get_anchor_id, based on url parameters,
        otherwise the first parameter will be assumed to be it"""
        return next(iter(kwargs.values()))

    async def list_instances(self, **kwargs):
        """Subclass must implement this if serializer has many=True parameter.
        Returns a queryset"""
        raise NotImplemented

    async def add_instance(self, instance_id, at_beginning=False):
        if instance_id in self.anchor_index:
            return
        if at_beginning:
            await self._consumer.prepend_anchor_id(instance_id)
        self.anchor_ids.append(instance_id)
        self.anchor_index.add(instance_id)
        await self._consumer.connect_anchor(instance_id)
        async with StateLoader(self, instance_id) as loader:
            async for instances in loader.list_instances():
                if instances:
                    for instance in instances:
                        instance['_operation'] = 'initial_state'
                    data = json_dumps(instances)
                    await self.send(text_data=data)

    async def set_runtime_var(self, var, value):
        self.runtime_state[var] = value
        payload = { 'runtimeVar': var, 'value': value }
        payload = json.dumps(payload)
        await self.send(text_data=payload)

    @database_sync_to_async
    def serialize_instance(self, instance, tstamp=0):
        return self._state_model.serialize_instance(instance, tstamp)

    @database_sync_to_async
    def _check_instance(self, qs, instance):
        instance = qs.filter(id=instance.id).first()
        return bool(instance)

    async def remove_instance(self, instance_id):
        return await self._remove_instance(instance_id, True)

    async def _remove_instance(self, instance_id, remove):
        if instance_id not in self.anchor_ids:
            return
        serialized = {
            'id': instance_id,
            '_operation': 'delete',
            '_instance_type': self._state_model.instance_type,
        }
        await self._consumer.disconnect_anchor(instance_id)
        await self.send(text_data=json.dumps([serialized], default=str))
        if remove:
            self.anchor_ids.remove(instance_id)
            self.anchor_index.remove(instance_id)

    async def clear(self):
        for instance_id in self.anchor_ids:
            await self._remove_instance(instance_id, False)
        self.anchor_ids = []
        self.anchor_index = set()

    @staticmethod
    def has_permission(user, **kwargs):
        """Implement this method to check if user has permission on a channel"""
        return NotImplemented

    async def is_visible(self, instance_id):
        """Implement this to check if a new instance should be added to this
        channel. You should check if user permission on instance"""
        return NotImplemented

    async def on_connect(self, tstamp):
        """Called after user has been authenticated.
        tstamp is the tstamp sent on connection, if this is a reconnection"""
        pass

    async def on_disconnect(self):
        """Called when user disconnects"""
        pass

    @classmethod
    async def clear_cache(cls, anchor_id):
        redis = RedisStateSession(cls, anchor_id)
        result = await redis.cooldown()
        if not result:
            return result
        await MongoStateSession.clear(cls, anchor_id)
        return result

    @classmethod
    def broadcast_instance(cls, anchor_id, instance, operation='update'):
        cls._signal_handler.broadcast_instance(anchor_id, instance, operation)

    @classmethod
    def broadcast_notification(cls, anchor_id, notification, user_id=None):
        notification['_instance_type'] = '_notification'
        cls._wsrouter.sync_dispatch([notification], anchor_id, user_id)



class UnknownProperty(Exception):
    pass

class AnchorDoesNotExist(Exception):
    pass

class UnauthorizedError(Exception):
    pass

class ForbiddenError(Exception):
    pass

class RxDjangoBug(Exception):
    pass

class ActionNotAsync(Exception):
    pass



from .exceptions import UnknownProperty


def is_related_property(model, property_name):
    key = _make_key(model, property_name)
    return key in RelatedProperty.accessors


def get_accessor(model, property_name):
    return RelatedProperty.get_accessor(model, property_name)


def get_reverse_accessor(model, property_name):
    return RelatedProperty.get_reverse_accessor(model, property_name)


class RelatedProperty:
    accessors = {}
    reverses = {}
    unknown_properties = set()

    def __init__(self, accessor, reverse):
        self.accessor = accessor
        self.reverse = reverse

    def __call__(self, fget):
        self.fget = fget
        key = (fget.__module__, fget.__qualname__)
        RelatedProperty.accessors[key] = self.accessor
        RelatedProperty.reverses[key] = self.reverse
        return property(self._get_wrapper())

    def _get_wrapper(self):
        def wrapper(instance):
            return self.fget(instance)
        return wrapper

    @classmethod
    def get_accessor(cls, model, property_name):
        key = _make_key(model, property_name)
        try:
            return cls.accessors[key]
        except KeyError:
            if key in cls.unknown_properties:
                name = '.'.join(key)
                module = cls.__module__
                raise UnknownProperty(
                    f'Unknown property {name}, use {module}.related_property '
                    'decorator instead to provide an acessor'
                )

    @classmethod
    def get_reverse_accessor(cls, model, property_name):
        key = _make_key(model, property_name)
        try:
            return cls.reverses[key]
        except KeyError:
            if key in cls.unknown_properties:
                name = '.'.join(key)
                module = cls.__module__
                raise UnknownProperty(
                    f'Unknown property {name}, use {module}.related_property '
                    'decorator instead to provide a reverse acessor'
                )

    @classmethod
    def register_unknown_property(cls, model, property_name):
        key = _make_key(model, property_name)
        cls.unknown_properties.add(key)
        return cls.accessors.get(key)

def _make_key(model, property_name):
    qualname = '.'.join([model.__name__, property_name])
    return (model.__module__, qualname)



#!/home/lfagundes/devel/rxdjango/rxdjango-env/bin/python

# $Id: rst2xetex.py 9115 2022-07-28 17:06:24Z milde $
# Author: Guenter Milde
# Copyright: This module has been placed in the public domain.

"""
A minimal front end to the Docutils Publisher, producing Lua/XeLaTeX code.
"""

try:
    import locale
    locale.setlocale(locale.LC_ALL, '')
except Exception:
    pass

from docutils.core import publish_cmdline

description = ('Generates LaTeX documents from standalone reStructuredText '
               'sources for compilation with the Unicode-aware TeX variants '
               'XeLaTeX or LuaLaTeX. '
               'Reads from <source> (default is stdin) and writes to '
               '<destination> (default is stdout).  See '
               '<https://docutils.sourceforge.io/docs/user/latex.html> for '
               'the full reference.')

publish_cmdline(writer_name='xetex', description=description)



import json
import redis
from asgiref.sync import async_to_sync
from django.utils.functional import cached_property
from django.conf import settings
from rxdjango.serialize import json_dumps


async def _connect():
    return await redis.asyncio.from_url(settings.REDIS_URL)

def _sync_connect():
    return redis.from_url(settings.REDIS_URL)


async def get_tstamp():
    conn = await _connect()
    tstamp = await conn.time()
    return _make_tstamp(tstamp)

def sync_get_tstamp():
    conn = _sync_connect()
    tstamp = conn.time()
    return _make_tstamp(tstamp)


def is_active(channel_class, anchor_id):
    key = _make_key(channel_class.name, anchor_id, 'state');
    conn = _sync_connect()
    # Return true if state is HEATING or HOT
    return conn.get(key) in (1, 2)


def _make_tstamp(tstamp):
    return tstamp[0] + tstamp[1] / 1000000

def _make_key(channel_name, anchor_id, key):
    return f'{channel_name}:{anchor_id}:{key}'



class RedisSession:

    # These are the variables for each anchor_id in Redis
    KEYS = [
        # The cache state - 0 COLD, 1 HEATING, 2 HOT or 3 COOLING
        'state',
        # The latest access time
        'access_time',
        # A list with all serialized instances, available
        # during HEATING and COOLING
        'instances',
        # The number of clients reading the list, to know when to cleanup
        'readers',
        # A pub/sub to notify on new instances as the list is built.
        # It's fed with the size of the list.
        'instances_trigger',
        # The number of clients relaying instances to the database
        'relay_writers',
        # A pub/sub to notify that relay_writers has changed
        'relay_writers_trigger',
    ]

    def __init__(self, channel, anchor_id):
        self.channel = channel

        self.local_keys = [
            _make_key(channel.name, anchor_id, key)
            for key in self.KEYS
        ]

        [
            self.state,
            self.access_time,
            self.instances,
            self.readers,
            self.instances_trigger,
            self.relay_writers,
            self.relay_writers_trigger,
        ] = self.local_keys

        self._conn = None

    async def connect(self):
        if self._conn:
            return self._conn
        self._conn = await _connect()
        return self._conn

    async def get_tstamp(self):
        tstamp = await self._conn.time()
        return _make_tstamp(tstamp)

    @classmethod
    def init_database(cls, channel_class):
        # This needs to be sync
        conn = _sync_connect()
        for key in conn.scan_iter(f"{channel_class.name}:*"):
            conn.delete(key)


class RedisStateSession(RedisSession):

    def __init__(self, channel, anchor_id):
        super().__init__(channel, anchor_id)
        self.initial_state = None
        self.tstamp = None
        self.written_instances = 0

    async def load(self):
        """Make an atomic call to redis, check the state and maybe transition

        States are:
          COLD
            - Documents are not cached.
            - Transition to HEATING
            - Clear the Redis instance list and set readers to 0
            - State come from ORM and go to Redis and Mongo
          HEATING
            - Documents are being cached
            - Increment readers
            - State come from Redis
          HOT
            - Documents are cached
            - State come from Mongo
          COOLING
            - Documents are being deleted from Mongo and availabe in Redis
            - Transition to HEATING
            - Increment readers
            - State come from Redis and go back go Mongo

        access_time is always updated

        Returns:
            int: The current state, represented as one of the class constants.
        """
        # 1 state,  2 access_time, 3 instances, 4 readers
        script = """
        local state = tonumber(redis.call("GET", KEYS[1])) or 0
        local tstamp = ARGV[1]

        if state == 0 then
            -- COLD state. Transition to HEATING
            redis.call("SET", KEYS[1], 1)

            -- clear instances list and reset readers
            redis.call("DEL", KEYS[3])
            redis.call("SET", KEYS[4], 0)

        elseif state == 1 then
            -- HEATING state. Increment readers
            redis.call("INCR", KEYS[4])

        elseif state == 3 then
            -- COOLING state. Transition to HEATING
            redis.call("SET", KEYS[1], 1)

            -- update readers to 1, as we're first reader
            redis.call("SET", KEYS[4], 1)
        end

        -- Update access timestamp
        redis.call("SET", KEYS[2], tstamp)

        return state
        """
        await self.connect()

        self.tstamp = await self.get_tstamp()

        start_session = self._conn.register_script(script)

        self.initial_state = await start_session(
            keys=self.local_keys,
            args=[self.tstamp],
        )

        return self.initial_state


    async def end_cold_session(self, success):
        """End a COLD session

        Transition state to HOT and delete the instances key if readers is 0.
        """
        if not success:
            return await self.rollback_to_cold()

        # 1 state,  2 access_time, 3 instances, 4 readers
        script = """
        local readers = tonumber(redis.call("GET", KEYS[4])) or 0

        if readers == 0 then
            -- delete instances key
            redis.call("DEL", KEYS[3])
        end

        -- transition to HOT
        redis.call("SET", KEYS[1], 2)

        return readers
        """
        await self.connect()
        end_cold_session = self._conn.register_script(script)
        return await end_cold_session(keys=self.local_keys)

    async def end_heating_session(self, success):
        """End a HEATING session

        Decrement readers and delete the instances key if readers reaches 0.
        """
        # 1 state,  2 access_time, 3 instances, 4 readers
        script = """
        local readers = tonumber(redis.call("DECR", KEYS[4])) or 0

        if readers == 0 then
            redis.call("DEL", KEYS[3]) -- delete instances key if readers is 0
        end

        return readers
        """

        await self.connect()
        end_heating_session = self._conn.register_script(script)
        return await end_heating_session(keys=self.local_keys)

    async def end_hot_session(self, success):
        """HOT session doesn't change anything"""
        pass

    async def end_cooling_session(self, success):
        """End COOLING session.

        Since we setup readers at 1, this is the same as end_heating_session.
        """
        await self.connect()
        return await self.end_heating_session(success)

    async def rollback_to_cold(self):
        """Transition to COLD, and broadcast error message to readers"""
        # 1 state,  2 access_time, 3 instances, 4 readers, 5 trigger, 6 lock, 7 lock_release
        script = """
        local readers = tonumber(redis.call("GET", KEYS[4])) or 0

        if readers > 0 then
            local error_message = 'error'
            redis.call("RPUSH", KEYS[3], error_message)
            local instances_size = redis.call("LLEN", KEYS[3]) -- Get the updated size of instances
            redis.call("PUBLISH", KEYS[5], instances_size) -- Publish the size to trigger channel
        end

        redis.call("SET", KEYS[1], 0) -- transition to COLD

        return readers
        """

        await self.connect()
        rollback_to_cold = self._conn.register_script(script)
        return await rollback_to_cold(keys=self.local_keys)

    async def write_instances(self, instances):
        """Serialize a list of objects, append them to the instances list, and publish the size to the trigger pub/sub channel.

        Args:
            objects (list): List of objects to append to the instances list.
        """
        serialized_instances = [json_dumps(instance) for instance in instances]

        await self.connect()
        start = self.written_instances

        for instance in instances:
            serialized = json_dumps(instance)
            await self._conn.rpush(self.instances, serialized)
            self.written_instances += 1

        await self._conn.publish(self.instances_trigger, self.written_instances)
        return self.written_instances

    async def end_write(self):
        """Notify all readers with a negative length, or delete instances if no readers"""

        script = """
        local readers = tonumber(redis.call("GET", KEYS[4])) or 0
        local negative_size = -tonumber(redis.call("LLEN", KEYS[3]))

        if readers == 0 then
            -- No readers, just clean instances
            redis.call("DEL", KEYS[3])
            return 0
        else
            -- Publish negative size to trigger
            redis.call("PUBLISH", KEYS[5], negative_size)
            return negative_size
        end
        """
        end_write = self._conn.register_script(script)
        return await end_write(keys=self.local_keys)

    async def list_instances(self):
        """An async generator that yields all instances available, then awaits on trigger and yields all new instances.

        When a negative value comes, it raises StopAsyncIteration after yielding the instances.
        """
        await self.connect()
        pubsub = self._conn.pubsub(ignore_subscribe_messages=True)
        await pubsub.subscribe(self.instances_trigger)
        # Get the initial length of instances and set up a cursor
        try:
            instances_length = await self._conn.llen(self.instances)
        except redis.exceptions.ResponseError:
            instances_length = 0
            await self._conn.delete(self.instances)

        last_length = 0
        cursor = 0

        while True:
            # If there are new instances, fetch and yield them using LRANGE
            if cursor < instances_length:
                new_instances = await self._conn.lrange(
                    self.instances, cursor, instances_length - 1
                )

                cursor = instances_length

                yield [json.loads(serialized) for serialized in new_instances]

            if last_length < 0:
                await pubsub.unsubscribe(self.instances_trigger)
                return

            message = await pubsub.get_message(timeout=5)
            if message is not None:
                last_length = int(message['data'])
                instances_length = abs(last_length)
            else:
                # Why do we get empty message?
                instances_length = await self._conn.llen(self.instances)
                if instances_length == last_length:
                    continue
                last_length = instances_length


    _end_session_methods = [
        end_cold_session,
        end_heating_session,
        end_hot_session,
        end_cooling_session,
    ]

    async def end(self, success):
        method = self._end_session_methods[self.initial_state]
        return await method(self, success)

    async def cooldown(self):
        # 1 state,  2 access_time, 3 instances, 4 readers
        script = """
        local state = tonumber(redis.call("GET", KEYS[1])) or 0

        if state == 0 then
            -- COLD state. Keep cold
            -- clear instances list and reset readers
            redis.call("DEL", KEYS[3])
            redis.call("SET", KEYS[4], 0)
            return 1

        elseif state == 1 then
            -- HEATING state. Let's not cooldown, keep heating
            return 0
        elseif state == 2 then
            -- HOT state. TODO proper transition to COOLING
            redis.call("SET", KEYS[1], 0)
            redis.call("DEL", KEYS[3])
            redis.call("SET", KEYS[4], 0)
            return 1
        elseif state == 3 then
            -- COOLING state. Let's not race it, keep cooling
            return 1
        else
            redis.call("SET", KEYS[1], 0)
            redis.call("DEL", KEYS[3])
            redis.call("SET", KEYS[4], 0)
            return 1
        end
        """
        await self.connect()
        cooldown = self._conn.register_script(script)
        result = await cooldown(
            keys=self.local_keys,
        )

        return result > 0



class _LockContextManager(RedisSession):
    """
    This is a base class for the two context managers that lock
    one channel for writing or cleanup.
    It manages the keys 6 and 7 and implements a semaphore to avoid
    cache cleanup while instances are being written.
    """
    timeout = 5

    def __init__(self, channel):
        super().__init__(channel)
        self.locked = None

    async def __aenter__(self):
        await self.connect()
        _, consumer = await self._conn.subscribe(self.relay_writers_trigger)

        acquire_write_lock = self._conn.register_script(self.lock_script)
        while True:
            acquired = await acquire_write_lock(keys=self.local_keys)

            if acquired >= 0:
                await self._conn.unsubscribe(self.relay_writers_trigger)
                self.locked = acquired > 0
                return

            if self.timeout is not None:
                try:
                    await asyncio.wait_for(consumer.get(), timeout=self.timeout)
                except asyncio.TimeoutError:
                    await self._conn.unsubscribe(self.relay_writers_trigger)
                    self.locked = False
                    return
            else:
                await consumer.get()

    async def __aexit__(self):
        if self.locked:
            release_write_lock = self._conn.register_script(self.release_script)
            await release_write_lock(keys=self.local_keys)


class WriteLock(_LockContextManager):
    """
    A non-exclusive lock on a channel to avoid cleanup while data is
    being written to mongo. Increments relay_writers (KEYS[6]) during
    write operation.
    """

    lock_script = """
        local state = tonumber(redis.call("GET", KEYS[1])) or 0

        if state == 0 or state == 3 then
            -- COLD or COOLING, no write should be done
            return 0
        end

        if state == 1 then
            -- HEATING, wait for finish
            return -1
        end

        local lock = tonumber(redis.call("GET", KEYS[6])) or 0
        if lock < 0 then
            -- Cleanup, wait for finish
            return -1
        end

        -- Increment lock
        redis.call("INCR", KEYS[6])
        redis.call("SET", KEYS[6], lock + 1) -- acquire lock

        return 1
        """

    release_script = """
        redis.call("DECR", KEYS[6])
        redis.call("PUBLISH", KEYS[7], 1)
        end
        """

class CleanupLock(_LockContextManager):
    """
    An exclusive lock on a channel to avoid new writes to mongo during
    cleanup. It sets relay_writers to negative number.
    """

    lock_script = """
        -- Set the lock key to 0 to publish to release pub/sub
        local writers = tonumber(redis.call("GET", KEYS[6])) or 0
        if writers < 0 then
            -- Another cleanup job running, do nothing
            return 0
        end
        if writers == 0 then
            redis.call("SET", KEYS[6], -1)
            redis.call("PUBLISH", KEYS[7], 1)
            return 1
        end

        -- Someone is writing, wait for release
        return -1
        """

    release_script = """
        redis.call("SET", KEYS[6], 0)
        redis.call("PUBLISH", KEYS[7], 1)
        end
        """



import os
from setuptools import find_packages
from skbuild import setup

here = os.path.abspath(os.path.dirname(__file__))

about = {}
with open(os.path.join(here, 'src', 'h3', '_version.py')) as f:
    exec(f.read(), about)


def long_desc():
    here = os.path.abspath(os.path.dirname(__file__))
    fname = os.path.join(here, 'readme.md')
    with open(fname) as f:
        long_description = f.read()

    return long_description


setup(
    name = 'h3',
    version = about['__version__'],
    description = about['__description__'],
    long_description = long_desc(),
    long_description_content_type = 'text/markdown',
    license = about['__license__'],
    author = about['__author__'],
    author_email = about['__author_email__'],
    url = about['__url__'],
    classifiers = about['__classifiers__'],
    include_package_data=True,
    packages = find_packages(
        'src',
        exclude = ["*.tests", "*.tests.*", "tests.*", "tests"],
    ),
    package_dir = {'': 'src'},
    cmake_languages = ('C'),
    extras_require={
        'numpy': ['numpy'],
        'test': ['pytest', 'pytest-cov', 'flake8', 'pylint'],
        'all': ['numpy', 'pytest', 'pytest-cov', 'flake8', 'pylint'],
    },
)



import sys
from django.core.management.base import BaseCommand
from rxdjango.sdk import make_sdk

class Command(BaseCommand):
    help=("Build the frontend SDK files."
          "Exit code 0 means code is up to date,"
          "1 means changes were made, or could me made (in case of dry-run)")


    def add_arguments(self, parser):
        parser.add_argument('--dry-run', action='store_true',
                            help="Do not write any changes")

        parser.add_argument('--quiet', action='store_true',
                            help="Do not output logs")
        parser.add_argument('--force', '-f', action='store_true',
                            help="Rebuild all files regardless of changes")


    def handle(self, *args, **kwargs):
        changed = make_sdk(
            not kwargs['dry_run'],
            kwargs['quiet'],
            kwargs['force'],
        )
        sys.exit(1 if changed else 0)



import json
from asgiref.sync import async_to_sync
import channels.layers
from rxdjango.serialize import json_dumps
from django.conf import settings


SYSTEM_CHANNEL = getattr(
    settings, 'RXDJANGO_SYSTEM_CHANNEL', '_rxdjango_system'
)


def get_channel_key(name, anchor_id, user_id=None):
    if user_id is None:
        return f'{name}_{anchor_id}'
    return f'{name}_{anchor_id}_{user_id}'


async def send_system_message(source, message):
    channel_layer = channels.layers.get_channel_layer()
    payload = {
        'source': source,
        'message': message,
    }
    await channel_layer.group_send(
        SYSTEM_CHANNEL,
        {
            'type': 'relay',
            'payload': payload,
        },
    )


class WebsocketRouter:

    def __init__(self, name):
        self.name = name

    async def connect(self, channel_layer, channel_name, anchor_id, user_id):
        # Join room group
        await channel_layer.group_add(
            get_channel_key(self.name, anchor_id),
            channel_name,
        )

        # Join this user room group, so that user-specific
        # data is shared
        await channel_layer.group_add(
            get_channel_key(self.name, anchor_id, user_id),
            channel_name,
        )

        # Connect to the system channel
        await channel_layer.group_add(SYSTEM_CHANNEL, channel_name)

    def sync_dispatch(self, payload, anchor_id, user_id=None):
        async_to_sync(self.dispatch)(payload, anchor_id, user_id)

    async def dispatch(self, payload, anchor_id, user_id=None):
        channel_key = get_channel_key(self.name, anchor_id, user_id)
        channel_layer = channels.layers.get_channel_layer()

        # FIXME: Datetime fields should be handled prior to this
        payload = json_dumps(payload)
        payload = json.loads(payload)

        await channel_layer.group_send(
            channel_key,
            {
                'type': 'relay',
                'payload': payload,
            },
        )

    async def disconnect(self, channel_layer, channel_name, anchor_id, user_id=None):
        # Leave the room group
        await channel_layer.group_discard(
            get_channel_key(self.name, anchor_id),
            channel_name,
        )

        # Leave this user room group
        if user_id is not None:
            await channel_layer.group_discard(
                get_channel_key(self.name, anchor_id, user_id),
                channel_name,
            )

        # Leave the system channel
        await channel_layer.group_discard(SYSTEM_CHANNEL, channel_name)



import json
from pytz import utc
from datetime import datetime


def default_serializer(value):
    if isinstance(value, datetime):
        value = utc.normalize(value)
        value = str(value).split('+')[0]
        return f'{value}Z'
    return str(value)


def json_dumps(value):    
    return json.dumps(value, default=default_serializer)



import json
from datetime import datetime
from copy import copy
from decimal import Decimal
import pymongo
import gridfs
from motor import motor_asyncio
from django.db import ProgrammingError
from django.conf import settings
from .redis import get_tstamp, sync_get_tstamp
from .serialize import json_dumps


class MongoStateSession:

    def __init__(self, channel, anchor_id):
        self.channel = channel
        self.anchor_id = anchor_id
        self.user_id = channel.user_id
        self.state_model = channel._state_model
        self._tstamp = None

        client = motor_asyncio.AsyncIOMotorClient(settings.MONGO_URL)
        self.db = client[settings.MONGO_STATE_DB]
        self.collection = self.db[channel.__class__.__name__.lower()]

    async def tstamp(self):
        if self._tstamp:
            return self._tstamp
        self._tstamp = await get_tstamp()
        return self._tstamp

    async def list_instances(self, user_id):
        for model in self.state_model.models():
            instances = []
            query = {
                '_anchor_id': self.anchor_id,
                '_user_key': {'$in': [None, user_id]},
                '_instance_type': model.instance_type,
                '_deleted': {'$ne': True},
            }

            async for instance in self.collection.find(query):
                try:
                    grid_ref = instance['_grid_ref']
                except KeyError:
                    pass
                else:
                    fs = gridfs.GridFS(self.db)
                    serialized = fs.get(grid_ref)
                    instance = json.loads(serialized.decode())

                del instance['_id']
                del instance['_anchor_id']
                instance['_operation'] = 'initial_state'
                instances.append(instance)

            if instances:
                yield instances
                instances = []

    async def write_instances(self, instances):
        for instance in instances:
            instance = _adapt(instance)
            instance['_anchor_id'] = self.anchor_id
            instance_type = instance['_instance_type']
            if instance.get('id', None) is None:
                raise ProgrammingError(f'Instance type {instance_type} has no "id"')
            instance = _adapt(instance)
            await self.collection.replace_one(
                {
                    '_anchor_id': self.anchor_id,
                    '_instance_type': instance['_instance_type'],
                    'id': instance['id'],
                },
                instance,
                upsert=True,
            )

    @staticmethod
    async def clear(channel_class, anchor_id):
        client = motor_asyncio.AsyncIOMotorClient(settings.MONGO_URL)
        db = client[settings.MONGO_STATE_DB]
        collection = db[channel_class.__name__.lower()]
        query = {'_anchor_id': anchor_id}
        await collection.delete_many(query)


class MongoSignalWriter:
    def __init__(self, channel_class):
        # Make a new connection, because this needs to be sync
        client = pymongo.MongoClient(settings.MONGO_URL)
        self.db = client[settings.MONGO_STATE_DB]
        self.collection = self.db[channel_class.__name__.lower()]

    def init_database(self):
        self.collection.drop()

        self.collection.create_index(
            [
                ('_anchor_id', pymongo.ASCENDING),
                ('_user_key', pymongo.ASCENDING),
                ('_instance_type', pymongo.ASCENDING),
                ('id', pymongo.ASCENDING),
            ],
            name='instance_pkey',
        )

        self.collection.create_index(
            [
                ('_anchor_id', pymongo.ASCENDING),
                ('_tstamp', pymongo.DESCENDING),
            ],
            name='reconnection_index',
        )

    def write_instances(self, anchor_id, instances):
        deltas = []

        for instance in instances:
            instance = _adapt(instance)
            instance['_anchor_id'] = anchor_id
            assert instance['_tstamp']
            try:
                original = self.collection.find_one_and_update(
                    {
                        '_anchor_id': anchor_id,
                        '_instance_type': instance['_instance_type'],
                        'id': instance['id'],
                    }, {
                        '$set': instance,
                    },
                    upsert=True,
                )
                if original:
                    del original['_id']
            except pymongo.errors.DocumentTooLarge:
                original = None
                data = json_dumps(instance).encode()
                fs = gridfs.GridFS(self.db)
                grid_ref = fs.put(data)

                instance = {
                    '_anchor_id': anchor_id,
                    '_instance_type': instance['_instance_type'],
                    'id': instance['id'],
                    '_grid_ref': grid_ref,
                }

                self.collection.replace_one(
                    {
                        '_anchor_id': anchor_id,
                        '_instance_type': instance['_instance_type'],
                        'id': instance['id'],
                    },
                    instance,
                    upsert=True,
                )

            if (original is None
                or instance['_operation'] == 'delete'
                or instance.get('_deleted', False) != original.get('_deleted', False)
                ):
                deltas.append(instance)
            else:
                empty = True
                for key, old_value in original.items():
                    if key == 'id' or key.startswith('_'):
                        continue
                    try:
                        new_value = instance[key]
                    except KeyError:
                        # An exception in a property may have generated
                        # an incomplete serialized object.
                        # TODO emit a warning
                        continue
                    if new_value == old_value:
                        del instance[key]
                    else:
                        empty = False

                if not empty:
                    deltas.append(instance)

        return deltas


def _adapt(instance):
    adapted = {}
    for key, value in instance.items():
        if isinstance(value, Decimal):
            value = float(value)
        elif isinstance(value, datetime):
            value = value.isoformat()[:26] + 'Z'
        adapted[key] = value

    return adapted



import asyncio
import typing
import inspect
from datetime import datetime
from collections import defaultdict
from .exceptions import ForbiddenError, ActionNotAsync

# A set of references for registered actions, so to protect non-action methods
# from being called by frontend
__actions = set()


def action(method):
    """Methods in a ContextChannel decorated with @action will be
    implemented in the frontend counterpart ContextChannel class,
    so that the method can be called directly from the frontend.
    """
    if not asyncio.iscoroutinefunction(method):
        raise ActionNotAsync(f'@action decorator requires "{method.__name__}" to be async')
    wrapped = method
    # Method may be decorated, find the original method ref
    while getattr(wrapped, '__wrapped__', None):
        wrapped = wrapped.__wrapped__
    # Register method to be callable
    __actions.add(wrapped)
    # Inspect method parameters so to make type conversions
    # when calling method
    hints = typing.get_type_hints(method)
    hints.pop('return', None)
    hints = list(hints.values())
    method.__datetime_fields = []
    for i in range(len(hints)):
        if hints[i] is datetime:
            method.__datetime_fields.append(i)
    # Return the original method
    return method


def list_actions(channel):
    """List all decorated methods in this django deployment"""
    for method in channel.__dict__.values():
        try:
            if method in __actions:
                yield method
        except TypeError:
            pass


async def execute_action(channel, method_name, params):
    method = getattr(channel, method_name, None)
    _verify_method(method)
    for i in method.__datetime_fields:
        params[i] = datetime.fromisoformat(params[i])
    return await method(*params)


def _verify_method(method):
    """Checks that a method is registered as an action"""
    if not method:
        raise ForbiddenError
    if getattr(method, '__func__', None):
        method = method.__func__
    while getattr(method, '__wrapped__', None):
        method = method.__wrapped__
    if not method in __actions:
        raise ForbiddenError



import os
import difflib
import subprocess
import types
import typing
from decimal import Decimal
from datetime import datetime
from django.db.models.query import QuerySet
from django.utils import timezone
from rest_framework import serializers

__serializers = set()

TYPEMAP = {
    int: 'number',
    float: 'number',
    Decimal: 'number',
    datetime: 'string',
    str: 'string',
    bool: 'boolean',
    type(None): 'null',
    QuerySet: 'number[]',
    # FIXME: we need to check the dict type and convert properly
    # Now we only support string: string
    dict: '{ [key: string]: string }',
}

def interface_name(Serializer):
    return Serializer.__name__.replace('Serializer', 'Type')


def export_interface(Serializer) -> None:
    """Mark this serializer so that its interface will be exported"""
    if issubclass(Serializer, serializers.Serializer):
        __serializers.add(_key(Serializer))


def ts_exported(Serializer) -> bool:
    """Check if a serializer interface should be exported as TS"""
    return _key(Serializer) in __serializers


def _key(Serializer):
    return '.'.join([Serializer.__module__, Serializer.__name__])

def header(app, *middle):
    now = timezone.now().strftime('%Y-%m-%d %H:%M')
    tz = timezone.now().tzinfo

    header = [
        f' * Generated by RxDjango on {now} ({tz})',
    ] + [
        f' * {line}' for line in middle
    ]

    footer = [
        f' * To rebuild it, run:',
        f' *     ./manage.py makefrontend [{app}]',
    ]

    warning = 'DO NOT EDIT THIS FILE'
    # make it fancy

    length = max([len(line) for line in header])
    w_left = int((length - len(warning)) / 2)
    w_right = length - len(warning) - w_left - 1

    header = [
        line + ' ' * (length - len(line) + 1) + '*'
        for line in header
    ]
    header.append(
        ' *' + ' ' * w_left + warning + ' ' * w_right + '*',
    )

    footer = [
        line + ' ' * (length - len(line) + 1) + '*'
        for line in footer
    ]

    code = [
        '/*' + '*' * length
    ]
    for line in header:
        code.append(line)
        code.append(
            ' *' + ' ' * (length - 1) + '*',
        )

    for line in footer:
        code.append(line)
        code.append(
            ' *' + ' ' * (length - 1) + '*',
        )

    code.append(
        ' ' + '*' * length + '*/'
    )

    return code


def diff(existing, content, filename):
    if len(existing) > 1:
        existing[0] = existing[1] = content[0] = content[1]

    filename = _git_relative_path(filename)

    fromfile = os.path.join('a', filename)
    tofile = os.path.join('b', filename)
    dif = difflib.unified_diff(existing, content, fromfile=fromfile, tofile=tofile)
    dif = list(dif)
    header, body = dif[:3], dif[3:]
    return ''.join(header) + '\n'.join(body)


def _git_relative_path(filepath):
    try:
        # Run the git command to get the relative path from the repository root
        git_root = subprocess.check_output(['git', 'rev-parse', '--show-toplevel']).strip().decode('utf-8')
        relative_path = os.path.relpath(filepath, git_root).strip()
        return relative_path
    except subprocess.CalledProcessError:
        # If the git command fails, return the original path
        return filepath

    
def get_ts_type(ftype):
    if type(ftype) is types.UnionType :
        py_types = typing.get_args(ftype)
    else:
        py_types = [ftype]

    ts_types = [ TYPEMAP[typ] for typ in py_types ]
    return ' | '.join(ts_types)


def snake_to_camel(s):
    components = s.split('_')
    # We capitalize the first letter of each component except the first one
    # with the 'title' method and join them together.
    return components[0] + ''.join(x.title() for x in components[1:])




import os
import re
import json
import typing
from collections import defaultdict
from channels.routing import ProtocolTypeRouter, URLRouter
from django.urls import URLPattern, URLResolver
from django.conf import settings
from rxdjango.consumers import StateConsumer
from rxdjango.actions import list_actions
from . import (header, interface_name, diff, get_ts_type, snake_to_camel,
               TYPEMAP)

def create_app_channels(app, apply_changes=True, force=False):
    consumer_urlpatterns = list_consumer_patterns(app)
    if not consumer_urlpatterns:
        return

    channel_module_name = f'{app}.channels'
    channel_path = channel_module_name.replace('.', '/') + '.py'

    ts_file_path = os.path.join(settings.RX_FRONTEND_DIR, f'{app}/{app}.channels.ts')
    py_mtime = None

    if os.path.exists(channel_path):
        py_mtime = os.path.getmtime(channel_path)

    existing = []

    if os.path.exists(ts_file_path):
        if not force and py_mtime == os.path.getmtime(ts_file_path):
            return
        with open(ts_file_path, 'r') as file:
            existing = file.read().split('\n')

    code = header(
        app,
        f'Based on all ContextChannel.as_asgi() calls in {settings.ASGI_APPLICATION}',
        f'This is expected to match {app}/channels.py',
    )

    code.extend([
        f"import {{ ContextChannel }} from '@rxdjango/react';\n",
        f'const SOCKET_URL = {settings.RX_WEBSOCKET_URL};',
    ])

    import_types = defaultdict(list)
    body = []

    for urlpattern in consumer_urlpatterns:
        consumer_class = urlpattern.callback.consumer_class
        context_channel_class = consumer_class.context_channel_class
        class_name = context_channel_class.__name__
        class_code = generate_ts_class(context_channel_class, urlpattern, import_types)
        body.append('\n')
        body.extend(class_code)

    if not body:
        return

    code.append('')  # line break
    code.extend(build_imports(import_types, app))
    code.extend(body)

    content = '\n'.join(code)

    if content.split('\n')[2:] == existing[2:]:
        if py_mtime:
            os.utime(ts_file_path, (py_mtime, py_mtime))
        return

    difference = diff(existing, content.split('\n'), ts_file_path)

    if not apply_changes:
        return difference

    try:
        with open(ts_file_path, 'w') as fh:
            fh.write(content)
    except FileNotFoundError:
        os.makedirs(os.path.dirname(ts_file_path), exist_ok=True)
        with open(ts_file_path, 'w') as fh:
            fh.write(content)

    if py_mtime:
        os.utime(ts_file_path, (py_mtime, py_mtime))

    return difference


def get_root_routing():
    asgi_app = settings.ASGI_APPLICATION
    module_name, app_name = asgi_app.rsplit('.', 1)
    module = __import__(module_name, fromlist=[app_name])
    return getattr(module, app_name)


def list_consumer_patterns(app_name, pattern_list=None, router=None):
    """
    Function to extract WebSocket consumers from the root routing (or provided router)
    for a specific app that extends the StateConsumer class.
    """

    if pattern_list is None:
        pattern_list = []
        router = get_root_routing()

    try:
        router = router.application
    except AttributeError:
        pass

    if isinstance(router, ProtocolTypeRouter):
        # Extract the websocket routing
        websocket_router = router.application_mapping.get('websocket', None)
        if websocket_router:
            list_consumer_patterns(app_name, pattern_list, websocket_router)

    elif isinstance(router, URLRouter):
        for route in router.routes:
            callback = route.callback
            if not hasattr(callback, 'consumer_class') or \
               not issubclass(callback.consumer_class, StateConsumer):
                continue

            context_channel_class =  callback.consumer_class.context_channel_class

            if context_channel_class.__module__.startswith(f'{app_name}.'):
                pattern_list.append(route)

            elif isinstance(route, URLRouter):
                # Recursive exploration for nested URLRouter objects
                list_consumer_patterns(app_name, pattern_list, route)

    return pattern_list


def pattern_to_ts(urlpattern):
    pattern_str = str(urlpattern.pattern)

    # Extract key-type pairs
    matches = re.findall(r'<(\w+?):(\w+?)>', pattern_str)

    # Basic mapping from Django path converters to TypeScript types
    type_mapping = {
        'str': 'string',
        'int': 'number',
        'slug': 'string',
        'uuid': 'string',
        'path': 'string',
        # Add more if needed
    }

    parameters = {key: type_mapping.get(type_, 'any') for type_, key in matches}

    # Convert Django pattern to a format easier to process in JS
    endpoint = re.sub(r'<(\w+?):(\w+?)>', r'{\2}', pattern_str)

    return endpoint, parameters

def build_imports(serializers, this_app):
    code = []
    for app, app_serializers in serializers.items():
        interfaces = [ interface_name(seri) for seri in app_serializers ]
        interfaces = ', '.join(sorted(interfaces))
        path = '.' if app == this_app else f'../{app}'
        code.append(f"import {{ {interfaces} }} from '{path}/{app}.interfaces.d';")

    return code

def generate_ts_class(context_channel_class, urlpattern, import_types):
    # First, we get the endpoint pattern and the parameters from our previous function
    endpoint, parameters = pattern_to_ts(urlpattern)

    name = context_channel_class.__name__
    anchor = context_channel_class.Meta.state
    if context_channel_class.many:
        many = 'true'
        anchor = anchor.child
    else:
        many = 'false'

    anchor_module = anchor.__class__.__module__
    anchor_name = anchor.__class__.__name__
    app = anchor_module.split('.')[0]
    state_type = interface_name(anchor.__class__)

    import_types[app].append(anchor.__class__)

    if context_channel_class.many:
        state_type += '[]'

    if getattr(context_channel_class, 'RuntimeState', False):
        runtime_type = f'{context_channel_class.__name__}RuntimeState'
        types = f'{state_type}, {runtime_type}'
    else:
        runtime_type = None
        types = state_type

    # Base of the class
    code = [
        f"export class {name} extends ContextChannel<{types}> {{\n",
        f"  anchor = '{anchor_module}.{anchor_name}';",
        f"  endpoint: string = '{endpoint}';",
        f"  args: {{ [key: string]: number | string }} = {{}};",
        f"  many = {many};",
    ]

    # Add private properties based on parameters
    # for key, ts_type in parameters.items():
    #     code.append(f"  {key}: {ts_type};")

    if runtime_type:
        code.append(f'  runtimeState: {runtime_type} | undefined')
        types = typing.get_type_hints(context_channel_class.RuntimeState)

        code = [
            f"export interface {runtime_type} {{"
        ] + [
            f"  {var}: {TYPEMAP[_type]};"
            for var, _type in types.items()
        ] + [
            "}\n"
        ] + code
    else:
        code.append(f'  runtimeState = null;')

    code.append(f'  baseURL: string = SOCKET_URL;\n')

    # Constructor
    params = ', '.join([f"{key}: {ts_type}"
                        for key, ts_type in parameters.items()])

    params = f'{params}, token: string' if params else 'token: string'

    code.append(f"  constructor({params}) {{")
    code.append(f"    super(token);")
    code.extend([f"    this.args['{key}'] = {key};" for key in parameters])
    code.append(f"  }}")

    code.append('')

    # Actions
    for action in list_actions(context_channel_class):
        hints = typing.get_type_hints(action)

        camel_action = snake_to_camel(action.__name__)

        return_type = hints.pop('return', type(None))
        return_type = get_ts_type(return_type)

        params = [ f'{k}: {get_ts_type(v)}' for k, v in hints.items() ]
        params = ', '.join(params)

        call_params = ', '.join(list(hints.keys()))

        code.append(f"  public async {camel_action}({params}): Promise<{return_type}> {{")
        code.append(f"    return await this.callAction('{action.__name__}', [{call_params}]);")
        code.append(f"  }}")

        code.append('')

    model = context_channel_class._state_model.frontend_model()
    model_code = json.dumps(model, indent=2)
    indented = "\n".join("  " + line for line in model_code.splitlines())
    code.append(f'  model = {indented[2:]};')

    code.append(f"}}")

    return code
